from __future__ import annotations

import os
import json
from pathlib import Path
from typing import TYPE_CHECKING

from quart import (
    Request,
    websocket,
    request,
    # copy_current_websocket_context,
)

from .defaults import Quart, logger
from .apis import GETS, POSTS, WS
from .data_manager import data_manager

if TYPE_CHECKING:
    from argparse import Namespace


def get_app(args: Namespace):
    """Get the Quart app."""

    class PipenBoardRequrest(Request):
        def __init__(self, *ags, **kwargs) -> None:
            super().__init__(*ags, **kwargs)
            self.cli_args = args

    os.environ["QUART_ENV"] = "development"
    app = Quart(
        __package__,
        static_folder=Path(__file__).parent / "frontend" / "build",
        static_url_path="/",
    )
    app.request_class = PipenBoardRequrest

    @app.after_request
    async def _(r):
        if args.dev:
            r.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
            r.headers["Pragma"] = "no-cache"
            r.headers["Expires"] = "0"
            r.headers["Cache-Control"] = "public, max-age=0"
        return r

    for route, handler in GETS.items():
        app.route(route, methods=["GET"])(handler)

    for route, handler in POSTS.items():
        app.route(route, methods=["POST"])(handler)

    clients = {}

    @app.websocket("/ws")
    async def ws():
        """The websocket handler"""
        client = None
        try:
            while True:
                message = await websocket.receive()
                message = json.loads(message)
                client = message["client"]
                if message["type"] == "connect":
                    clients[client] = websocket._get_current_object()
                    await WS[f"{client}/conn"](clients)
                else:
                    await WS[client](message, clients)
        finally:
            await WS[f"{client}/disconn"](clients)

    @app.route("/api/run", methods=["POST"])
    async def run():
        data = await request.get_json()

        command = data["command"]
        overwriteConfig = data["overwriteConfig"]
        config = data["config"]
        tomlfile = Path(data["tomlfile"])
        if not overwriteConfig and tomlfile.exists():
            return {
                "ok": False,
                "msg": (
                    "Config file already exists. "
                    "Check the box to overwrite it, "
                    "or use a different configuration file name."
                ),
            }
        try:
            tomlfile.write_text(config)
        except Exception as ex:
            return {
                "ok": False,
                "msg": f"Failed to write the configuration file: {ex}",
            }

        if data_manager.running:
            return {
                "ok": False,
                "msg": (
                    "Pipeline is already running. "
                    "Please wait for it to finish, or stop it first."
                ),
            }

        logger.info(
            f"[bold][yellow]API[/yellow][/bold] Running pipeline: {command}"
        )
        # Run command at background
        app.add_background_task(
            data_manager.run_pipeline,
            command,
            args.port,
            clients,
        )
        return {"ok": True, "msg": ""}

    @app.route("/api/pipeline/rerun", methods=["POST"])
    async def rerun():
        logger.info(
            f"[bold][yellow]API[/yellow][/bold] Re-Running pipeline: %s",
            data_manager._command,
        )

        if not data_manager._command:
            return {
                "ok": False,
                "error": "Pipeline never ran. Please reload the page.",
            }

        # Run command at background
        app.add_background_task(
            data_manager.run_pipeline,
            data_manager._command,
            args.port,
            clients,
        )
        return {"ok": True}

    return app



from __future__ import annotations

import base64
import json
import re
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from quart import abort, request, redirect, send_file
from slugify import slugify

from .version import __version__
from .defaults import (
    JOB_STATUS,
    PIPEN_BOARD_DIR,
    SECTION_PIPELINE_OPTIONS,
    logger,
)
from .data_manager import data_manager


if TYPE_CHECKING:
    from typing import Mapping, Any, Tuple


# Helper functions
def _get_children(parent: Path, idx: int = 0) -> Tuple[Mapping[str, Any], int]:
    """Get the children of a parent path"""
    out = []
    for child in parent.glob("*"):
        idx += 1
        item = {"id": idx, "text": child.name, "full": str(child)}
        if child.is_dir():
            item["children"], idx = _get_children(child, idx)
        out.append(item)
    return out, idx


def _get_file_content(path: Path, how: str) -> str:
    how, n = how.split(" ", 1)
    n = int(n)
    lines = path.read_text().splitlines()

    if how == "Head":
        return "\n".join(lines[:n])

    return "\n".join(lines[-n:])


def _is_text_file(file_path: str | Path) -> bool:
    """Check if the file is a text file"""
    try:
        with open(file_path) as file:
            file.read(1)
    except UnicodeDecodeError:
        return False
    return True


# APIs
async def index():
    """Redirect to the index.html"""
    if request.cli_args.dev:
        return redirect("index.html?dev=1")
    return redirect("index.html")


async def history():
    logger.info("[bold][yellow]API[/yellow][/bold] Getting histories")
    args = request.cli_args
    out = {}
    out["pipeline"] = args.pipeline
    out["histories"] = []
    curr_workdir = Path(args.workdir).resolve()

    for histfile in PIPEN_BOARD_DIR.glob(f"{slugify(args.pipeline)}.*.*.json"):
        name = histfile.stem.split(".")[-2]
        workdir = base64.b64decode(
            histfile.stem.split(".")[-1] + "=="
        ).decode()
        out["histories"].append(
            {
                "name": name,
                "configfile": histfile.name,
                "workdir": workdir,
                "is_current": Path(workdir).resolve() == curr_workdir,
                # 2023-01-01_00-00-00 to
                # 2023-01-01 00:00:00
                "ctime": (
                    datetime.fromtimestamp(histfile.stat().st_ctime).strftime(
                        "%Y-%m-%d %H:%M:%S"
                    )
                ),
                "mtime": (
                    datetime.fromtimestamp(histfile.stat().st_mtime).strftime(
                        "%Y-%m-%d %H:%M:%S"
                    )
                ),
            }
        )

    return out


async def version():
    """Get the version of pipen-board"""
    return __version__


async def pipeline_data():
    logger.info("[bold][yellow]API[/yellow][/bold] Getting pipeline data")
    req = await request.get_json()
    configfile = req.get("configfile")
    preset = req.get("preset")
    return data_manager.get_data(request.cli_args, configfile, preset)


async def reports(report_path):
    """Get the reports"""
    root, rest = report_path.split("/", 1)
    root = root.replace("|", "/")

    report_path = Path(root) / rest
    if report_path.is_dir():
        report_path = report_path / "index.html"

    if not report_path.is_file():
        return abort(404)
    # Serve the file
    return await send_file(report_path)


async def report_building_log():
    """Get the building log of a report"""
    args = request.cli_args
    name = request.args["name"]
    report_file = Path(args.workdir).joinpath(
        name,
        ".report-workdir",
        "pipen-report.log",
    )
    logger.info(
        "[bold][yellow]API[/yellow][/bold] Getting building log: %s",
        report_file,
    )
    if not report_file.is_file():
        return {"ok": False, "content": "No report building log file found."}

    pattern = re.compile(r"\x1B\[\d+(;\d+){0,2}m")
    return {"ok": True, "content": pattern.sub("", report_file.read_text())}


async def history_del():
    configfile = (await request.get_json())["configfile"]
    logger.info(
        "[bold][yellow]API[/yellow][/bold] Deleting history: %s",
        configfile,
    )
    PIPEN_BOARD_DIR.joinpath(configfile).unlink()
    return {"ok": True}


async def history_saveas():
    req = await request.get_json()
    args = request.cli_args
    configfile = req["configfile"]
    newname = req["new_name"]

    logger.info(
        "[bold][yellow]API[/yellow][/bold] Save history: %s with new name: %s",
        configfile,
        newname,
    )
    try:
        jdata = json.loads(PIPEN_BOARD_DIR.joinpath(configfile).read_text())
        jdata[SECTION_PIPELINE_OPTIONS]["name"]["value"] = newname

        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        workdir = Path(args.workdir).resolve().as_posix()
        out = {
            "name": newname,
            "is_current": True,
            "mtime": now,
            "ctime": now,
            "workdir": workdir,
        }
        enc = base64.b64encode(workdir.encode()).decode().rstrip("=")
        newconfigfile = PIPEN_BOARD_DIR.joinpath(
            f"{slugify(args.pipeline)}.{newname}.{enc}.json"
        )

        if newconfigfile.is_file():
            return {"ok": False, "error": "File already exists."}

        for val in jdata.get("RUNNING_OPTIONS", {}).values():
            configfile_opt = val.get("configfile", "configfile")
            if not val["value"][configfile_opt].get("changed"):
                val["value"][configfile_opt]["value"] = val["value"][
                    configfile_opt
                ]["placeholder"] = f"{newname}.config.toml"

        newconfigfile.write_text(json.dumps(jdata, indent=4))
        out["configfile"] = newconfigfile.name

    except Exception as exc:
        return {"ok": False, "error": str(exc)}
    return out


async def history_download():
    req = await request.get_json()
    configfile = req["configfile"]
    logger.info(
        "[bold][yellow]API[/yellow][/bold] Downloading schema: %s",
        configfile,
    )
    return await send_file(
        PIPEN_BOARD_DIR.joinpath(configfile),
        as_attachment=True,
    )


# async def history_upload():
#     # get form data
#     form = await request.files
#     # load as json
#     jdata = json.load(form["schema_file"])
#     name = jdata[SECTION_PIPELINE_OPTIONS]["name"]["value"]
#     logger.info(
#         "[bold][yellow]API[/yellow][/bold] Receiving schema file with name: "
#         f"{name}"
#     )
#     workdir = Path(request.cli_args.workdir).resolve().as_posix()
#     enc = base64.b64encode(workdir.encode()).decode().rstrip("=")
#     schema_file = PIPEN_BOARD_DIR.joinpath(
#         f"{slugify(request.cli_args.pipeline)}.{name}.{enc}.json"
#     )
#     if schema_file.is_file():
#         return {"ok": False, "error": "File already exists."}

#     schema_file.write_text(json.dumps(jdata, indent=4))
#     return {
#         "name": name,
#         "mtime": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
#         "ctime": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
#         "workdir": workdir,
#         "is_current": True,
#         "configfile": schema_file.name,
#     }


async def history_fromurl():
    # get form data
    url = (await request.get_json())["url"]
    logger.info(
        "[bold][yellow]API[/yellow][/bold] Receiving TOML file from url: %s",
        url,
    )
    try:
        import urllib.request
        with urllib.request.urlopen(url) as response:
            return {"ok": True, "content": response.read().decode()}

    except Exception as exc:
        return {"ok": False, "error": str(exc)}


async def config_save():
    args = request.cli_args
    data = await request.get_json()
    configfile = data.get("configfile")
    configdata = data["data"]
    jdata = json.loads(configdata)

    now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    name = jdata[SECTION_PIPELINE_OPTIONS]["name"]["value"]
    workdir = Path(args.workdir).resolve().as_posix()
    out = {"name": name, "mtime": now}
    enc = base64.b64encode(workdir.encode()).decode().rstrip("=")
    if not configfile:
        # base64 encode the workdir path
        configfile = PIPEN_BOARD_DIR.joinpath(
            f"{slugify(args.pipeline)}.{name}.{enc}.json"
        )
        for val in jdata.get("RUNNING_OPTIONS", {}).values():
            configfile_opt = val.get("configfile", "configfile")
            if not val["value"][configfile_opt].get("changed"):
                val["value"][configfile_opt]["value"] = val["value"][
                    configfile_opt
                ]["placeholder"] = f"{name}.config.toml"

        out["ctime"] = now
        out["is_current"] = True
        out["workdir"] = workdir
        logger.info(
            "[bold][yellow]API[/yellow][/bold] Saving config to a new file: "
            f"{configfile}"
        )
    elif configfile.startswith("new:"):
        name = (
            configfile[4:]
            or jdata[SECTION_PIPELINE_OPTIONS]["name"]["default"]
        )
        jdata[SECTION_PIPELINE_OPTIONS]["name"]["value"] = name
        for val in jdata.get("RUNNING_OPTIONS", {}).values():
            configfile_opt = val.get("configfile", "configfile")
            if not val["value"][configfile_opt].get("changed"):
                val["value"][configfile_opt]["value"] = val["value"][
                    configfile_opt
                ]["placeholder"] = f"{name}.config.toml"

        configdata = json.dumps(jdata, indent=4)
        configfile = PIPEN_BOARD_DIR.joinpath(
            f"{slugify(args.pipeline)}.{name}.{enc}.json"
        )
        if configfile.exists():
            return {"ok": False, "error": "File already exists."}

        out["name"] = name
        out["ctime"] = now
        out["is_current"] = True
        out["workdir"] = workdir
        logger.info(
            "[bold][yellow]API[/yellow][/bold] Saving config to a new file: "
            f"{configfile}"
        )
    else:
        configfile = PIPEN_BOARD_DIR.joinpath(configfile)
        logger.info(
            f"[bold][yellow]API[/yellow][/bold] Saving config to: {configfile}"
        )

    out["configfile"] = configfile.name
    configfile.write_text(configdata)
    return out


async def job_get_tree():
    args = request.cli_args
    data = await request.get_json()
    logger.info(
        "[bold][yellow]API[/yellow][/bold] Fetching tree for: "
        f"{data['proc']}/{data['job']}"
    )
    jobdir = Path(args.workdir).joinpath(
        data["name"],
        data["proc"],
        str(data["job"]),
    )
    if not jobdir.is_dir():
        return []

    # compose a treeview data
    # see: https://carbon-components-svelte.onrender.com/components/TreeView
    return _get_children(jobdir)[0]


async def job_get_file():
    """Get file details"""
    data = await request.get_json()
    how = data.get("how", "full")
    path = Path(data["path"])
    if how != "full":
        logger.info(
            "[bold][yellow]API[/yellow][/bold] Fetching file for "
            f"{data['proc']}/{data['job']}: {path} ({how})"
        )
        return {
            "type": "bigtext-part",
            "content": _get_file_content(path, how),
        }

    logger.info(
        "[bold][yellow]API[/yellow][/bold] Fetching file for "
        f"{data['proc']}/{data['job']}: {path}"
    )
    if path.name.startswith("job.wrapped.") or path.name in (
        "job.script",
        "job.signature.toml",
        "job.rc",
    ):
        return {"type": "text", "content": path.read_text()}

    if path.name == "job.status":
        st_code = path.read_text().strip()
        status = JOB_STATUS.get(st_code, "UNKNOWN")
        return {"type": "text", "content": f"{st_code} ({status})"}

    # check different types of files and sizes of them
    if path.suffix in (".png", ".jpg", ".jpeg", ".gif", ".svg"):
        return {
            "type": "image",
            "content": (
                f"data:image/{path.suffix[1:]};base64,"
                f"{base64.b64encode(path.read_bytes()).decode()}"
            ),
        }

    if _is_text_file(path):
        # check the size of the file
        size = path.stat().st_size
        if size > 1024 * 1024:  # 1MB
            # read first 100 lines
            lines = []
            with path.open() as f:
                for i, line in enumerate(f):
                    lines.append(line)
                    if i > 100:
                        break
            return {"type": "bigtext", "content": "".join(lines)}

        return {"type": "text", "content": path.read_text()}

    return {"type": "binary"}


async def job_get_file_metadata():
    data = await request.get_json()
    path = Path(data["path"])
    logger.info(
        "[bold][yellow]API[/yellow][/bold] Fetching file metadata for "
        f"{data['proc']}/{data['job']}: {path}"
    )

    size = size_human = path.stat().st_size
    base = 1024
    size_human = abs(float(size_human))
    for unit in ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB"]:
        if size_human < base:
            break
        size_human /= base

    return {
        "name": path.name,
        "size": size,
        "size_human": f"{size_human:.2f} {unit}",
        "ctime": datetime.fromtimestamp(path.stat().st_ctime).isoformat(
            sep=" ", timespec="seconds"
        ),
        "mtime": datetime.fromtimestamp(path.stat().st_mtime).isoformat(
            sep=" ", timespec="seconds"
        ),
    }


async def pipeline_stop():
    return await data_manager.stop_pipeline()


async def ws_web(data, clients):
    logger.info(f"WS/WEB Received: {data}")


async def ws_pipeline(data, clients):
    # logdata = str(data)
    # if len(logdata) > 100:
    #     logdata = logdata[:100] + "..."

    # logger.info(f"WS/PIPELINE Received: {logdata}")
    type = data.get("type")
    if (
        type
        and type.startswith("on_")
        and callable(getattr(data_manager, type, None))
    ):
        await getattr(data_manager, type)(data["data"], clients.get("web"))


async def ws_web_conn(clients):
    logger.info("WS/WEB Client 'web' connected.")
    # send the current run data, to let UI know the current status
    await data_manager.send_run_data(clients["web"], True)


async def ws_pipeline_conn(clients):
    logger.info("WS/PIPELINE Client 'pipeline' connected.")


async def ws_web_disconn(clients):
    logger.info("WS/WEB Client 'web' disconnected.")


async def ws_pipeline_disconn(clients):
    logger.info("WS/PIPELINE Client 'pipeline' disconnected.")
    data_manager.running = False


GETS = {
    "/": index,
    "/api/history": history,
    "/api/version": version,
    "/api/report_building_log": report_building_log,
    "/reports/<path:report_path>": reports,
}

POSTS = {
    "/api/pipeline": pipeline_data,
    "/api/history/del": history_del,
    "/api/history/saveas": history_saveas,
    "/api/history/download": history_download,
    # "/api/history/upload": history_upload,
    "/api/history/fromurl": history_fromurl,
    "/api/config/save": config_save,
    "/api/job/get_tree": job_get_tree,
    "/api/job/get_file": job_get_file,
    "/api/job/get_file_metadata": job_get_file_metadata,
    "/api/pipeline/stop": pipeline_stop,
}

WS = {
    "web": ws_web,
    "pipeline": ws_pipeline,
    "web/conn": ws_web_conn,
    "pipeline/conn": ws_pipeline_conn,
    "web/disconn": ws_web_disconn,
    "pipeline/disconn": ws_pipeline_disconn,
}



from __future__ import annotations

import logging
import json
import sys
import selectors
from typing import TYPE_CHECKING

import websocket
from pipen.utils import get_marked, get_logger
from pipen.pluginmgr import plugin

from .version import __version__
from .defaults import (
    NAME,
    SECTION_PROCESSES,
    SECTION_PROCGROUPS,
    SECTION_DIAGRAM,
    SECTION_REPORTS,
)

if TYPE_CHECKING:
    from pipen import Pipen, Proc
    from pipen.job import Job

logger = get_logger(NAME)


class PipenBoardPlugin:
    name = NAME
    # Let other plugins run first
    priority = 9999
    __version__ = __version__

    def __init__(self):
        self.ws = None

    def _send(self, data, log=None):
        if self.ws:
            data["client"] = "pipeline"
            try:
                self.ws.send(json.dumps(data))
            except BrokenPipeError:
                pass

            logdata = str(data)
            if len(logdata) > 100:
                logdata = logdata[:100] + "..."
            logmsg = f"SENDING {logdata}"
            if log is None:
                logger.debug(logmsg)
            else:
                log("debug", logmsg, logger=logger)

    def _connect(self):
        """Connect to pipen-board server"""
        if not sys.stdin.readable():
            self.ws = None
            logger.debug("Stdin is not readable, skip.")
            return

        sel = selectors.DefaultSelector()
        try:
            # nohup fails
            sel.register(sys.stdin, selectors.EVENT_READ)
        except PermissionError:
            self.ws = None
            logger.debug("No permission to access selectors, skip.")
            return

        if not sel.select(timeout=1):
            self.ws = None
            logger.debug("No data in stdin, skip.")
            return

        port = sys.stdin.readline().strip()
        if not port.startswith("pipen-board:"):
            self.ws = None
            logger.debug("Not spawned by pipen-board server, skip.")
            return

        # Now that we are spawned by pipen-board
        port = int(port[12:])
        self.ws = websocket.WebSocket()
        self.ws.connect(f"ws://localhost:{port}/ws")
        logger.info(f"Connected to pipen-board at ws://localhost:{port}/ws")
        self._send({"type": "connect", "client": "pipeline"})

    def _disconnect(self):
        if self.ws:
            self.ws.close()
            self.ws = None

    @plugin.impl
    async def on_start(self, pipen: Pipen):
        logger.setLevel(
            getattr(logging, pipen.config.get("loglevel", "INFO").upper())
        )
        self._connect()
        if not self.ws:
            return

        data = {}
        diagram = pipen.outdir.joinpath("diagram.svg")
        if diagram.is_file():
            data[SECTION_DIAGRAM] = diagram.read_text()

        for proc in pipen.procs:
            pg = get_marked(proc, "procgroup")
            if pg:
                group = data.setdefault(SECTION_PROCGROUPS, {})
                group.setdefault(pg.name, []).append(proc.name)
            else:
                data.setdefault(SECTION_PROCESSES, []).append(proc.name)

        self._send({"type": "on_start", "data": data})

    @plugin.impl
    async def on_complete(self, pipen: Pipen, succeeded: bool):
        data = {"succeeded": succeeded}
        if succeeded:
            reports_dir = pipen.outdir.joinpath("REPORTS")
            if (
                reports_dir.joinpath("index.html").is_file()
                and reports_dir.joinpath("procs").is_dir()
                and [p for p in reports_dir.joinpath("procs").iterdir()]
            ):
                data[SECTION_REPORTS] = str(reports_dir.parent)

        self._send({"type": "on_complete", "data": data})
        self._disconnect()

    @plugin.impl
    async def on_proc_start(self, proc: Proc):
        group = get_marked(proc, "procgroup")
        if group:
            group = group.name

        self._send(
            {
                "type": "on_proc_start",
                "data": {
                    "proc": proc.name,
                    "procgroup": group,
                    "njobs": len(proc.jobs),
                },
            },
            log=proc.log,
        )

    @plugin.impl
    async def on_proc_done(self, proc: Proc, succeeded: bool):
        group = get_marked(proc, "procgroup")
        if group:
            group = group.name

        self._send(
            {
                "type": "on_proc_done",
                "data": {
                    "proc": proc.name,
                    "procgroup": group,
                    "succeeded": succeeded,
                },
            },
            log=proc.log,
        )

    @plugin.impl
    async def on_job_queued(self, proc: Proc, job: Job):
        group = get_marked(proc, "procgroup")
        if group:
            group = group.name

        self._send(
            {
                "type": "on_job_queued",
                "data": {
                    "proc": proc.name,
                    "procgroup": group,
                    "job": job.index,
                },
            },
            log=job.log,
        )

    @plugin.impl
    async def on_job_submitted(self, proc: Proc, job: Job):
        group = get_marked(proc, "procgroup")
        if group:
            group = group.name

        self._send(
            {
                "type": "on_job_submitted",
                "data": {
                    "proc": proc.name,
                    "procgroup": group,
                    "job": job.index,
                },
            },
            log=job.log,
        )

    @plugin.impl
    async def on_job_running(self, proc: Proc, job: Job):
        group = get_marked(proc, "procgroup")
        if group:
            group = group.name

        self._send(
            {
                "type": "on_job_running",
                "data": {
                    "proc": proc.name,
                    "procgroup": group,
                    "job": job.index,
                },
            },
            log=job.log,
        )

    @plugin.impl
    async def on_job_killed(self, proc: Proc, job: Job):
        group = get_marked(proc, "procgroup")
        if group:
            group = group.name

        self._send(
            {
                "type": "on_job_killed",
                "data": {
                    "proc": proc.name,
                    "procgroup": group,
                    "job": job.index,
                },
            },
            log=job.log,
        )

    @plugin.impl
    async def on_job_succeeded(self, proc: Proc, job: Job):
        group = get_marked(proc, "procgroup")
        if group:
            group = group.name

        self._send(
            {
                "type": "on_job_succeeded",
                "data": {
                    "proc": proc.name,
                    "procgroup": group,
                    "job": job.index,
                },
            },
            log=job.log,
        )

    @plugin.impl
    async def on_job_failed(self, proc: Proc, job: Job):
        group = get_marked(proc, "procgroup")
        if group:
            group = group.name

        self._send(
            {
                "type": "on_job_failed",
                "data": {
                    "proc": proc.name,
                    "procgroup": group,
                    "job": job.index,
                },
            },
            log=job.log,
        )

    @plugin.impl
    async def on_job_cached(self, proc: Proc, job: Job):
        group = get_marked(proc, "procgroup")
        if group:
            group = group.name

        self._send(
            {
                "type": "on_job_cached",
                "data": {
                    "proc": proc.name,
                    "procgroup": group,
                    "job": job.index,
                },
            },
            log=job.log,
        )



"""Provides the running data and functions to mutate it"""
from __future__ import annotations

import asyncio
import json
import os
import re
import signal
import sys
import time
import importlib
import importlib.util
from copy import deepcopy
from multiprocessing import Process, Pipe
from pathlib import Path
from tempfile import gettempdir
from typing import TYPE_CHECKING, Any, Mapping, Type
from urllib.parse import urlparse

from simpleconf import Config
from slugify import slugify
from liquid import Liquid
from pipen import Pipen, Proc, ProcGroup
from pipen.utils import get_marked
from pipen_annotate import annotate

from .defaults import (
    PIPEN_BOARD_DIR,
    SECTION_PIPELINE_OPTIONS,
    SECTION_ADDITIONAL_OPTIONS,
    SECTION_PROCESSES,
    SECTION_PROCGROUPS,
    SECTION_DIAGRAM,
    SECTION_REPORTS,
    SECTION_LOG,
    PIPELINE_OPTIONS,
    logger,
)

if TYPE_CHECKING:
    from argparse import Namespace

DEFAULT_RUN_DATA = {
    "FINISHED": False,
    # "log": None
    SECTION_LOG: None,
    # "diagram": "<svg>...</svg>"
    SECTION_DIAGRAM: None,
    # "reports": <reportdir>
    SECTION_REPORTS: None,
    # "PROCESSES": { proc: { status, jobs: [status] } }
    SECTION_PROCESSES: {},
    # "PROCGROUPS": {
    #     procgroup: { proc: { status, jobs: [status] } }
    # }
    SECTION_PROCGROUPS: {},
}

# proc status: init, running, succeeded, failed
# job status: init, queued, submitted, running, killed, succeeded, failed


def parse_pipeline(pipeline: str) -> Pipen:
    """Parse the pipeline"""
    modpath, sep, name = pipeline.rpartition(":")
    if sep != ":":
        raise ValueError(
            f"Invalid pipeline: {pipeline}.\n"
            "It must be in the format '<module[.submodule]>:pipeline' or \n"
            "'/path/to/pipeline.py:pipeline'"
        )

    path = Path(modpath)
    if path.is_file():
        spec = importlib.util.spec_from_file_location(path.stem, modpath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    else:
        module = importlib.import_module(modpath)

    try:
        pipeline = getattr(module, name)
    except AttributeError:
        raise ValueError(f"Invalid pipeline: {pipeline}") from None

    if isinstance(pipeline, type) and issubclass(pipeline, Proc):
        pipeline = Pipen(name=f"{pipeline.name}Pipeline").set_starts(pipeline)

    if isinstance(pipeline, type) and issubclass(pipeline, Pipen):
        # Avoid "pipeline" to be used as pipeline name
        [pipeline] = [pipeline()]

    if isinstance(pipeline, type) and issubclass(pipeline, ProcGroup):
        pipeline = pipeline().as_pipen()

    if not isinstance(pipeline, Pipen):
        raise ValueError(
            f"Invalid pipeline: {pipeline}\n"
            "It must be a `pipen.Pipen` instance"
        )

    return pipeline


def _anno_to_argspec(anno: Mapping[str, Any] | None) -> Mapping[str, Any]:
    """Convert the annotation to the argument spec"""
    if anno is None:
        return {}

    argspec = {}
    # arginfo: attrs, help, terms
    for arg, arginfo in anno.items():
        argspec[arg] = arginfo.attrs.copy()
        argspec[arg]["desc"] = re.sub(
            r"([\.\?!:])\s*\n",
            "\\1<br />\n",
            arginfo.help,
        )
        if arg.startswith("<") and arg.endswith(">"):
            argspec[arg].setdefault("order", 999)
        if "btype" not in argspec[arg]:
            if (
                argspec[arg].get("text")
                or argspec[arg].get("mline")
                or argspec[arg].get("mlines")
            ):
                argspec[arg]["type"] = "text"
            elif argspec[arg].get("action") in (
                "store_true",
                "store_false",
            ) or argspec[arg].get("flag"):
                argspec[arg]["type"] = "bool"
            elif (
                argspec[arg].get("action") in ("ns", "namespace")
                or argspec[arg].get("ns")
                or argspec[arg].get("namespace")
            ):
                argspec[arg]["type"] = "ns"
            elif (
                argspec[arg].get("action")
                in ("append", "extend", "clear_append", "clear_extend")
                or argspec[arg].get("array")
                or argspec[arg].get("list")
            ):
                argspec[arg]["type"] = "list"
            elif argspec[arg].get("choices") or argspec[arg].get("choice"):
                argspec[arg]["type"] = "choice"
            elif argspec[arg].get("mchoice") or argspec[arg].get("mchoices"):
                argspec[arg]["type"] = "mchoice"
        else:
            argspec[arg]["type"] = argspec[arg].pop("btype")

        t = argspec[arg].get("type")
        if t == "ns":
            argspec[arg]["value"] = _anno_to_argspec(arginfo.terms)
        elif t in ("choice", "mchoice"):
            argspec[arg]["value"] = argspec[arg].get("default", [])
            argspec[arg]["choices"] = list(arginfo.terms)
            argspec[arg]["choices_desc"] = [
                term.help if not term.help else term.help.splitlines()[0]
                for term in arginfo.terms.values()
            ]
            argspec[arg]["desc"] += "\n"
            # Add the choices to the description
            for termname, term in arginfo.terms.items():
                h = re.sub(r"([\.\?!:])\s*\n", "\\1<br />\n", term.help)
                argspec[arg]["desc"] += f"- `{termname}`: {h}\n"
        else:
            argspec[arg]["value"] = argspec[arg].get("default", None)

        # determine the itype for list elements
        if t == "list":
            if argspec[arg]["value"] is not None and not isinstance(
                argspec[arg]["value"], list
            ):
                argspec[arg]["value"] = [argspec[arg]["value"]]
            if (
                argspec[arg]["value"] is not None
                and argspec[arg]["value"]
                and "bitype" not in argspec[arg]
                and "itype" not in argspec[arg]
                and not isinstance(argspec[arg]["value"][0], str)
            ):
                argspec[arg]["itype"] = type(argspec[arg]["value"][0]).__name__

            if "bitype" in argspec[arg]:
                argspec[arg]["itype"] = argspec[arg].pop("bitype")

    # argspec.setdefault("default", argspec.get("value", None))
    return argspec


def _get_default(argspec: Mapping[str, Any], force_ns: bool = False) -> Any:
    """Set the default using the value for the argspec

    This can be used to omit the argument in final config in frontend
    in case the value is the same as the default
    """
    value = argspec.get("value", None)
    if not force_ns and argspec.get("type") not in ("ns", "namespace"):
        argspec.setdefault("default", value)
        return argspec["default"]

    # flatten the namespace
    if value is None:
        argspec.setdefault("default", {})
        return argspec["default"]

    default = {}
    for arg, arginfo in value.items():
        # placeholder
        if arg.startswith("<") and arg.endswith(">"):
            continue

        # Set the default for sub-item
        dlt = _get_default(arginfo)
        default[arg] = dlt

    argspec.setdefault("default", default)
    return default


def _proc_to_argspec(
    proc: Proc | Type[Proc],
    is_start: bool,
    hidden: bool = False,
    order: int = 0,
) -> Mapping[str, Any]:
    """Convert the proc to the argument spec"""
    if isinstance(proc, Proc):
        anno = annotate(proc.__class__)
    else:
        anno = annotate(proc)

    summary = anno.get("Summary", {"short": "", "long": ""})
    argspec = {
        "is_start": is_start,
        "order": order,
        "desc": f'# {summary["short"]}\n\n{summary["long"]}',
        "value": {},
    }
    if hidden:
        argspec["hidden"] = True
        return argspec

    if is_start and not get_marked(proc, "board_config_no_input"):
        argspec["value"]["in"] = {
            "desc": "The input data for the process",
            "type": "ns",
            "required": True,
            "value": _anno_to_argspec(anno.get("Input", {})),
        }
    argspec["value"]["envs"] = {
        "desc": "Environment variables for the process, used across jobs",
        "value": _anno_to_argspec(anno.get("Envs", {})),
    }
    _get_default(argspec["value"]["envs"], force_ns=True)

    argspec["value"]["plugin_opts"] = PIPELINE_OPTIONS["plugin_opts"]
    _get_default(argspec["value"]["plugin_opts"], force_ns=True)

    argspec["value"]["scheduler_opts"] = PIPELINE_OPTIONS["scheduler_opts"]
    _get_default(argspec["value"]["scheduler_opts"], force_ns=True)

    argspec["value"]["forks"] = PIPELINE_OPTIONS["forks"]
    _get_default(argspec["value"]["forks"])

    argspec["value"]["cache"] = PIPELINE_OPTIONS["cache"]
    _get_default(argspec["value"]["cache"])

    argspec["value"]["scheduler"] = PIPELINE_OPTIONS["scheduler"]
    _get_default(argspec["value"]["scheduler"])

    argspec["value"]["dirsig"] = PIPELINE_OPTIONS["dirsig"]
    _get_default(argspec["value"]["dirsig"])

    argspec["value"]["error_strategy"] = PIPELINE_OPTIONS["error_strategy"]
    _get_default(argspec["value"]["error_strategy"])

    argspec["value"]["num_retries"] = PIPELINE_OPTIONS["num_retries"]
    _get_default(argspec["value"]["num_retries"])

    argspec["value"]["lang"] = {
        "desc": "The interpreter to run the script",
        "hidden": True,
        "placeholder": proc.lang,
        # Don't include it in the config file if not specified
        "value": None,
        "default": None,
    }

    return argspec


def _load_additional(additional: str, **kwargs) -> Mapping[str, Any]:
    """Load additional config files

    Args:
        additional: The additional config file to load
        kwargs: Key-value pairs to render the additional config file,
            which is a liquid/jinja2 template
    """
    if not additional:
        return {}

    if additional.startswith("gh://"):
        # To avoid the first part being parsed as hostname
        additional = f"gh:{additional[5:]}"

    parsed = urlparse(additional)
    cache_dir = Path(gettempdir()) / "pipen-cli-config-additional-configs"
    cache_dir.mkdir(parents=True, exist_ok=True)
    if parsed.scheme in ("http", "https", "ftp", "ftps", "gh"):
        from hashlib import sha256
        from urllib.error import URLError
        from urllib.request import urlretrieve

        if parsed.scheme == "gh":
            try:
                user, repo, file_path = parsed.path.split("/", 2)
            except ValueError:
                raise ValueError(f"Invalid gh:// URL: {additional}")
            branch = "master"
            if "@" in file_path:
                file_path, branch = file_path.split("@")
            parsed = urlparse(
                "https://raw.githubusercontent.com/"
                f"{user}/{repo}/{branch}/{file_path}"
            )

        url = parsed.geturl()
        cache_key = sha256(url.encode("utf-8")).hexdigest()
        additional = cache_dir / f"{cache_key}-{parsed.path.split('/')[-1]}"
        if not additional.exists():
            try:
                urlretrieve(url, additional)
            except URLError:
                raise ValueError(
                    f"Could not retrieve remote path: {additional}"
                ) from None

    if kwargs:
        # kwargs passed, treat the file as a template
        additional = Path(additional)
        tpl = Liquid(additional.read_text(), mode="wild", from_file=False)
        additional = cache_dir.joinpath(
            f"{slugify(str(additional.resolve()))}.rendered{additional.suffix}"
        )
        additional.write_text(tpl.render(**kwargs))

    out = Config.load(additional)
    if "ADDITIONAL_OPTIONS" in out:
        for val in out["ADDITIONAL_OPTIONS"].values():
            _get_default(val)

    return out


def _update_dict(d1: Mapping[str, Any], d2: Mapping[str, Any]) -> None:
    """Update d1 with d2 recursively"""
    for key, val in d2.items():
        if key in d1 and isinstance(val, dict) and isinstance(d1[key], dict):
            _update_dict(d1[key], val)
        else:
            d1[key] = val


async def _get_config_data(
    args: Namespace,
    name: str | None,
) -> Mapping[str, Any]:
    """Get the pipeline data"""
    try:
        old_argv = sys.argv
        sys.argv = ["@pipen-board"] + args.pipeline_args
        logger.info("[bold][yellow]DBG[/yellow][/bold] Fetching pipeline data ...")
        try:
            pipeline = parse_pipeline(args.pipeline)
            # Initialize the pipeline so that the arguments definied by
            # other plugins (i.e. pipen-args) to take in place.
            pipeline.workdir = Path(pipeline.config.workdir).joinpath(
                name or pipeline.name
            )
            await pipeline._init()
            pipeline.workdir.mkdir(parents=True, exist_ok=True)
            pipeline.build_proc_relationships()
        finally:
            sys.argv = old_argv

        data = {}
        data[SECTION_PIPELINE_OPTIONS] = PIPELINE_OPTIONS.copy()
        data[SECTION_PIPELINE_OPTIONS]["name"] = {
            "type": "str",
            "value": name or pipeline.name,
            "default": pipeline.name,
            "placeholder": name or pipeline.name,
            "readonly": True,
            "desc": (
                "The name of the pipeline. "
                "It will affect the names of working directory and "
                "the result directory"
            ),
        }
        data[SECTION_PIPELINE_OPTIONS]["desc"] = {
            "type": "str",
            "value": pipeline.desc,
            "desc": (
                "The description of the pipeline, " "shows in the log and report."
            ),
        }
        data[SECTION_PIPELINE_OPTIONS]["outdir"] = {
            "desc": "The output directory of your pipeline",
            "placeholder": "./<name>-output",
            "type": "str",
            "value": None,
        }
        for key, val in data[SECTION_PIPELINE_OPTIONS].items():
            _get_default(val, key in ("plugin_opts", "scheduler_opts"))

        data[SECTION_PROCESSES] = {}

        if pipeline.config.plugin_opts.get("args_flatten") is True or (
            "args_flatten" not in pipeline.config.plugin_opts
            and len(pipeline.procs) == 1
        ):
            data[SECTION_PIPELINE_OPTIONS]["plugin_opts"]["value"][
                "args_flatten"
            ] = {
                "desc": (
                    "Flatten the arguments of the pipeline. "
                    "For example, [envs] will ba treated as [<Process>.envs]. "
                    "Only works for single-process pipeline"
                ),
                "type": "bool",
                "value": True,
                "default": True,
                "readonly": True,
            }

        pg_sec = {}
        for i, proc in enumerate(pipeline.procs):
            logger.info(
                "[bold][yellow]DBG[/yellow][/bold] Parsing process: %s ...",
                proc,
            )

            pg = proc.__meta__["procgroup"]
            if pg:
                if pg.name not in pg_sec:
                    pg_sec[pg.name] = {"PROCESSES": {}}
                    pg_anno = annotate(pg.__class__)
                    # desc
                    pg_summ = pg_anno.get("Summary", {"short": "", "long": ""})
                    pg_sec[pg.name][
                        "desc"
                    ] = f'# {pg_summ["short"]}\n\n{pg_summ["long"]}'
                    # args
                    pg_args = _anno_to_argspec(pg_anno.get("Args")) or {}
                    # Implemented by pipen-annotate 0.7.3
                    # for arg, arginfo in pg_args.items():
                    #     arginfo["value"] = pg.DEFAULTS.get(arg)
                    pg_sec[pg.name]["ARGUMENTS"] = pg_args
                pg_sec[pg.name][SECTION_PROCESSES][proc.name] = _proc_to_argspec(
                    proc,
                    proc in pipeline.starts,
                    get_marked(proc, "board_config_hidden", False),
                    order=i,
                )
            else:
                data[SECTION_PROCESSES][proc.name] = _proc_to_argspec(
                    proc,
                    proc in pipeline.starts,
                    get_marked(proc, "board_config_hidden", False),
                    order=i,
                )

        data[SECTION_PROCGROUPS] = pg_sec

        if args.additional:
            logger.info(
                "[bold][yellow]DBG[/yellow][/bold] "
                "Loading additional configuration items ..."
            )
            if (
                args.additional == "auto"
                and args.pipeline.rpartition(":")[0].endswith(".py")
            ):
                additional = str(
                    Path(__file__).parent.joinpath("additional_auto.toml")
                )
            else:
                additional = args.additional

            addi_data = _load_additional(
                additional,
                name=name or pipeline.name,
                pipeline=args.pipeline,
                pipeline_args=args.pipeline_args,
            )
            _update_dict(data, addi_data)

    except Exception as e:
        import traceback
        return {"error": traceback.format_exc()}

    return data


class DataManager:
    """Gather and manager the pipeline data"""

    # Send data every 5 seconds to the client, via websocket
    INTERVAL = 5

    def __init__(self) -> None:
        self.running: int | bool = False
        self._config_data = None
        self._run_data = None
        self._timer = None
        self._proc_running_order = 0
        self._command = None

    def _get_config_data(
        self,
        args: Namespace,
        configfile: str,
    ):
        """Get the config data of the pipeline

        Args:
            args: The arguments from the CLI
            configfile: The name to the config file
        """
        if configfile and not configfile.startswith("new:") and not args.dev:
            with PIPEN_BOARD_DIR.joinpath(configfile).open() as f:
                self._config_data = json.load(f)
                return

        self._config_data = None
        if not configfile:
            name = None
        elif configfile.startswith("new:"):
            name = configfile[4:]
        else:
            self._config_data = json.loads(
                PIPEN_BOARD_DIR.joinpath(configfile).read_text()
            )
            name = self._config_data[SECTION_PIPELINE_OPTIONS]["name"]["value"]

        if not self._config_data:
            # Use multiprocessing to get a clean environment
            # to load the pipeline to avoid conflicts
            def target(conn):
                conn.send(
                    json.dumps(
                        asyncio.run(_get_config_data(args, name))
                    ).encode()
                )
                conn.close()

            parent_conn, child_conn = Pipe()
            p = Process(target=target, args=(child_conn,))
            p.start()
            data = parent_conn.recv()
            p.join()

            self._config_data = json.loads(data)
            if "error" in self._config_data:
                for line in self._config_data["error"].splitlines():
                    logger.error(line)

                from quart import abort
                abort(500)

        self._config_data[SECTION_PIPELINE_OPTIONS]["name"]["value"] = (
            name
            or self._config_data[SECTION_PIPELINE_OPTIONS]["name"]["default"]
        )

    def _get_prev_run(self, args: Namespace, configfile: str | None):
        """Get data for the previous run

        Args:
            args: The parsed arguments from cli
            configfile: The config file to use, if not specified, use a
                fresh one loaded from the pipeline object
                It could also be "new:<name>" for a new instance.
        """
        out = self._run_data = {SECTION_LOG: None, "FINISHED": True}
        self._get_config_data(args, configfile=configfile)
        name = self._config_data[SECTION_PIPELINE_OPTIONS]["name"]["value"]
        pipeline_dir = Path(args.workdir).joinpath(name)
        if not pipeline_dir.is_dir():
            # no previous run, return defaults
            self._run_data = DEFAULT_RUN_DATA
            return

        # Get the log
        logfile = pipeline_dir.joinpath("run-latest.log")
        if logfile.exists() and logfile.is_file():
            out[SECTION_LOG] = logfile.read_text()
        else:
            # no previous run, return defaults
            self._run_data = DEFAULT_RUN_DATA
            return

        config_data = self._config_data
        outdir = config_data[SECTION_PIPELINE_OPTIONS].get(
            "outdir",
            {"value": None},
        )["value"]
        if not outdir:
            outdir = Path(args.workdir).parent.joinpath(f"{name}-output")
        else:
            outdir = Path(outdir)

        report_procs_dir = outdir.joinpath("REPORTS", "procs")
        if report_procs_dir.is_dir() and [
            p for p in report_procs_dir.glob("*") if p.is_dir()
        ]:
            out[SECTION_REPORTS] = str(outdir)

        diagram = outdir.joinpath("diagram.svg")
        if diagram.is_file():
            out[SECTION_DIAGRAM] = diagram.read_text()

        out[SECTION_PROCESSES] = {}
        out[SECTION_PROCGROUPS] = {}

        def process_proc(proc: str, container: dict):
            procdir = pipeline_dir.joinpath(proc)
            container[proc] = {"jobs": []}
            if not procdir.is_dir():
                container[proc]["status"] = "init"
                return

            for jobdir in sorted(
                [j for j in procdir.glob("*") if j.name.isdigit()],
                key=lambda x: int(x.name),
            ):
                rcfile = jobdir / "job.rc"
                try:
                    rc = int(rcfile.read_text().strip())
                except Exception:
                    container[proc]["jobs"].append("failed")
                else:
                    container[proc]["jobs"].append(
                        "succeeded" if rc == 0 else "failed"
                    )
            if not container[proc]["jobs"]:
                container[proc]["status"] = "init"
            elif "failed" in container[proc]["jobs"]:
                container[proc]["status"] = "failed"
            else:
                container[proc]["status"] = "succeeded"

        for proc in config_data[SECTION_PROCESSES]:
            process_proc(proc, out[SECTION_PROCESSES])

        for pg in config_data[SECTION_PROCGROUPS]:
            out[SECTION_PROCGROUPS][pg] = {}
            for proc in config_data[SECTION_PROCGROUPS][pg][SECTION_PROCESSES]:
                process_proc(proc, out[SECTION_PROCGROUPS][pg])

    def _update_config_by_preset(self, preset: Mapping[str, Any] | None):
        if not preset:
            return

        def update_value(key, value, preset_val, force_ns=False):
            if not preset_val or key not in preset_val:
                return

            vtype = value.get("type")

            if vtype == "ns" or vtype == "namespace" or force_ns:
                for k, v in value["value"].items():
                    update_value(k, v, preset_val[key])
                for k, v in preset_val[key].items():
                    if k not in value["value"]:
                        value["value"][k] = {"value" : v}
            else:
                value["value"] = preset_val[key]
                if value.get("pgarg"):
                    value["changed"] = True

        # fill up the default value of self._config_data with preset
        if SECTION_PIPELINE_OPTIONS in self._config_data:
            for key, val in self._config_data[SECTION_PIPELINE_OPTIONS].items():
                update_value(key, val, preset, force_ns=key.endswith("_opts"))

        if SECTION_ADDITIONAL_OPTIONS in self._config_data:
            for key, val in self._config_data[SECTION_ADDITIONAL_OPTIONS].items():
                update_value(key, val, preset)

        if SECTION_PROCESSES in self._config_data:
            for proc, procconfig in self._config_data[SECTION_PROCESSES].items():
                for key, val in procconfig["value"].items():
                    update_value(
                        key, val, preset.get(proc), force_ns=key == "envs"
                    )

        if SECTION_PROCGROUPS in self._config_data:
            for pg, pgconfig in self._config_data[SECTION_PROCGROUPS].items():
                if "ARGUMENTS" in pgconfig:
                    for key, val in pgconfig["ARGUMENTS"].items():
                        update_value(key, val, preset.get(pg))
                if "PROCESSES" in pgconfig:
                    for proc, procconfig in pgconfig["PROCESSES"].items():
                        for key, val in procconfig["value"].items():
                            update_value(
                                key,
                                val,
                                preset.get(proc),
                                force_ns=key == "envs"
                            )


    def clear_run_data(self, keep_log: bool = False):
        """Clear the data"""
        self._proc_running_order = 0
        if not keep_log:
            self._run_data = deepcopy(DEFAULT_RUN_DATA)
        else:
            log = self._run_data[SECTION_LOG]
            self._run_data = deepcopy(DEFAULT_RUN_DATA)
            self._run_data[SECTION_LOG] = log

    def get_data(
        self,
        args: Namespace,
        configfile: str | None,
        preset: Mapping[str, Any] | None,
    ):
        """Get the data"""
        if not self.running:
            self._get_prev_run(args, configfile=configfile)
            self._update_config_by_preset(preset)
            return {
                "runStarted": False,
                "config": self._config_data,
                "run": self._run_data,
            }

        self._get_config_data(args, configfile=configfile)
        self._update_config_by_preset(preset)
        return {
            "runStarted": True,
            "config": self._config_data,
            "run": self._run_data,
        }

    async def send_run_data(self, ws, force: bool = False):
        if ws is None:
            return
        if not force and time.time() - self._timer < self.INTERVAL:
            return

        # Send data
        logger.debug(
            "[bold][yellow]DBG[/yellow][/bold] Sending run data to the frontend"
        )
        try:
            await ws.send(json.dumps(self._run_data))
        except BrokenPipeError:
            pass

        # Reset timer
        self._timer = time.time()

    async def on_start(self, data, ws):
        self._proc_running_order = 0
        # { SECTION_PROCESSES: [p1, p2], SECTION_PROCGROUPS: {pg1: [p3, p4]} }
        if isinstance(data, str):
            data = json.loads(data)

        logger.info("WS/PIPELINE Received: Pipeline started")

        if SECTION_DIAGRAM in data:
            self._run_data[SECTION_DIAGRAM] = data[SECTION_DIAGRAM]

        if SECTION_PROCESSES in data:
            for proc in data[SECTION_PROCESSES]:
                self._run_data[SECTION_PROCESSES][proc] = {
                    "order": 0,
                    "status": "init",
                    "jobs": [],
                }

        if SECTION_PROCGROUPS in data:
            for pg in data[SECTION_PROCGROUPS]:
                self._run_data[SECTION_PROCGROUPS][pg] = {}
                for proc in data[SECTION_PROCGROUPS][pg]:
                    self._run_data[SECTION_PROCGROUPS][pg][proc] = {
                        "status": "init",
                        "jobs": [],
                    }

        self.timer = time.time()
        await self.send_run_data(ws, force=True)

    async def on_complete(self, data, ws):
        if isinstance(data, str):
            data = json.loads(data)

        logger.info(
            "WS/PIPELINE Received: Pipeline completed (succeeded=%s)",
            data['succeeded'],
        )

        if SECTION_REPORTS in data:
            self._run_data[SECTION_REPORTS] = data[SECTION_REPORTS]

        self.running = False
        self._run_data["FINISHED"] = True
        await self.send_run_data(ws, force=True)
        self._proc_running_order = 0

    async def on_proc_start(self, data, ws):
        if isinstance(data, str):
            data = json.loads(data)

        proc, group, njobs = data["proc"], data["procgroup"], data["njobs"]

        logger.info(
            "WS/PIPELINE Received: Process started: %s (size=%s)",
            proc,
            njobs,
        )

        if not group:
            self._run_data[SECTION_PROCESSES][proc][
                "order"
            ] = self._proc_running_order
            self._run_data[SECTION_PROCESSES][proc]["status"] = "running"
            self._run_data[SECTION_PROCESSES][proc]["jobs"] = ["init"] * njobs
        else:
            self._run_data[SECTION_PROCGROUPS][group][proc][
                "order"
            ] = self._proc_running_order
            self._run_data[SECTION_PROCGROUPS][group][proc][
                "status"
            ] = "running"
            self._run_data[SECTION_PROCGROUPS][group][proc]["jobs"] = [
                "init"
            ] * njobs

        self._proc_running_order += 1
        await self.send_run_data(ws, force=True)

    async def on_proc_done(self, data, ws):
        if isinstance(data, str):
            data = json.loads(data)

        proc, group, succeeded = (
            data["proc"],
            data["procgroup"],
            data["succeeded"],
        )

        logger.info(
            "WS/PIPELINE Received: Process done: %s (succeeded=%s)",
            proc,
            succeeded,
        )

        procdata = (
            self._run_data[SECTION_PROCESSES][proc]
            if not group
            else self._run_data[SECTION_PROCGROUPS][group][proc]
        )
        # succeeded could be True, False, or "cached"
        procdata["status"] = "succeeded" if succeeded else "failed"

        await self.send_run_data(ws, force=True)

    async def _on_job(self, data, status, ws):
        if isinstance(data, str):
            data = json.loads(data)

        proc, group, job = data["proc"], data["procgroup"], data["job"]

        logger.info(
            "WS/PIPELINE Received: Job %s (%s#%s)",
            status,
            proc,
            job,
        )

        if not group:
            self._run_data[SECTION_PROCESSES][proc]["jobs"][job] = status
        else:
            self._run_data[SECTION_PROCGROUPS][group][proc]["jobs"][
                job
            ] = status
        await self.send_run_data(ws)

    async def on_job_queued(self, data, ws):
        await self._on_job(data, "queued", ws)

    async def on_job_submitted(self, data, ws):
        await self._on_job(data, "submitted", ws)

    async def on_job_running(self, data, ws):
        await self._on_job(data, "running", ws)
        # Notify the frontend that the job is running
        await self.send_run_data(ws, force=True)

    async def on_job_killed(self, data, ws):
        await self._on_job(data, "killed", ws)

    async def on_job_failed(self, data, ws):
        await self._on_job(data, "failed", ws)

    async def on_job_succeeded(self, data, ws):
        await self._on_job(data, "succeeded", ws)

    async def on_job_cached(self, data, ws):
        await self._on_job(data, "succeeded", ws)

    async def run_pipeline(self, command, port, ws_clients):
        """Run a command and send the output to the websocket"""
        self.clear_run_data()
        self._run_data[SECTION_LOG] = self._run_data[SECTION_LOG] or ""

        p = await asyncio.create_subprocess_shell(
            command,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )
        try:
            p.stdin.write(f"pipen-board:{port}\n".encode())
            p.stdin.close()
        except RuntimeError:
            # command is already finished, probably due to an error
            pass
        self.running = p.pid
        self._command = command

        while True:
            line = await p.stdout.readline()
            if not line:
                break

            self._run_data[SECTION_LOG] += line.decode()
            # In case it's too long to data between hooks
            await self.send_run_data(ws_clients.get("web"))

        if await p.wait() != 0:
            # In case the pipeline fails to start
            self._run_data["FINISHED"] = "error"
        else:
            self._run_data["FINISHED"] = True

        self.running = False
        await self.send_run_data(ws_clients.get("web"), force=True)

    async def stop_pipeline(self):
        """Stop the pipeline"""
        if not self.running:
            return {"ok": False, "msg": "Pipeline is not running"}

        logger.debug(
            "[bold][yellow]DBG[/yellow][/bold] Killing pipeline at %s ...",
            self.running,
        )

        # Let the pipeline send signals to the jobs
        # The jobs could be on some scheduler systems
        try:
            os.kill(self.running, signal.SIGINT)
        except ProcessLookupError:
            return {
                "ok": False,
                "msg": (
                    "Pipeline probably failed to start. You may want to "
                    "restart the pipen-board server and try again."
                ),
            }

        await asyncio.sleep(3)
        # Start killing
        import psutil

        try:
            proc = psutil.Process(self.running)
        except psutil.NoSuchProcess:
            # Might be killed by SIGINT
            pass
        else:
            for child in proc.children(recursive=True):
                try:
                    child.terminate()
                    await asyncio.sleep(1)
                except psutil.NoSuchProcess:
                    pass
                try:
                    child.kill()
                    await asyncio.sleep(1)
                except psutil.NoSuchProcess:
                    pass

            try:
                proc.terminate()
                await asyncio.sleep(1)
                proc.kill()
                await asyncio.sleep(1)
            except psutil.NoSuchProcess:
                pass

        self.running = False
        self._run_data["FINISHED"] = "error"
        return {"ok": True, "msg": "Pipeline killed"}


data_manager = DataManager()



"""Provides PipenCliConfigPlugin"""
from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from pipen.cli import CLIPlugin

from .version import __version__
from .defaults import NAME, logger
from .quart_app import get_app

if TYPE_CHECKING:  # pragma: no cover
    from argx import ArgumentParser, Namespace


class PipenCliBoardPlugin(CLIPlugin):
    """Configure and run pipen pipelines from the web"""

    name = NAME
    __version__ = __version__

    def __init__(
        self,
        parser: ArgumentParser,
        subparser: ArgumentParser,
    ) -> None:
        super().__init__(parser, subparser)
        subparser.usage = "%(prog)s [options] <pipeline> -- [pipeline options]"
        subparser.add_argument(
            "-p",
            "--port",
            type=int,
            default=18521,
            help="Port to serve the UI wizard",
        )
        subparser.add_argument(
            "-a",
            "--additional",
            metavar="FILE",
            help=(
                "Additional arguments for the pipeline, "
                "in YAML, INI, JSON or TOML format. "
                "Can have sections `ADDITIONAL_OPTIONS` and `RUNNING_OPTIONS`. "
                "It can also have other sections and items to override the "
                "configurations generated from the pipeline. "
                "If the pipeline is provided as a python script, such as "
                "`/path/to/pipeline.py:<pipeline>`, and `<pipeline>` runs "
                "under `__name__ == '__main__'`, the additional file can also "
                "be specified as `auto` to generate a `RUNNING OPTIONS/Local` "
                "section to run the pipeline locally."
            ),
        )
        subparser.add_argument(
            "--loglevel",
            default="auto",
            choices=["auto", "debug", "info", "warning", "error", "critical"],
            help=(
                "The logging level. If `auto`, it will be set to `debug` "
                "if `--dev` is set, otherwise `info`."
            ),
        )
        subparser.add_argument(
            "--dev",
            action="store_true",
            help=(
                "Run the pipeline in development/debug mode. "
                "This will reload the server when changes are made to this "
                "package and reload the pipeline when page reloads for new "
                "configurations. Page cache is also disabled in this mode."
            ),
        )
        subparser.add_argument(
            "-w",
            "--workdir",
            help="The working directory of the pipeline.",
            default=".pipen",
        )
        subparser.add_argument(
            "pipeline",
            help=(
                "The pipeline and the CLI arguments to run the pipeline. "
                "For the pipeline either `/path/to/pipeline.py:<pipeline>` "
                "or `<module.submodule>:<pipeline>` "
                "`<pipeline>` must be an instance of `Pipen` and running "
                "the pipeline should be called under `__name__ == '__main__'."
            ),
        )

    def parse_args(self) -> Namespace:
        """Parse the arguments"""
        # split the args into two parts, separated by `--`
        # the first part is the args for pipen_cli_config
        # the second part is the args for the pipeline
        args = sys.argv[1:]
        idx = args.index("--") if "--" in args else len(args)
        args, rest = args[:idx], args[idx + 1 :]
        parsed = self.parser.parse_args(args=args)
        parsed.pipeline_args = rest
        return parsed

    def exec_command(self, args: Namespace) -> None:
        """Execute the command"""
        if args.loglevel == "auto":
            logger.setLevel("DEBUG" if args.dev else "INFO")
        else:
            logger.setLevel(args.loglevel.upper())

        print(" * ")
        print(" *        __   __  __         __  __      __  __")
        print(" *       |__)||__)|_ |\ | __ |__)/  \ /\ |__)|  \\")
        print(" *       |   ||   |__| \|    |__)\__//--\| \ |__/")
        print(" * ")
        print(" *                   version: %s" % __version__)
        print(" * ")
        print("\n".join(map(lambda x: f" * {x}", self.__doc__.splitlines())))
        print(" * ")

        app = get_app(args)
        # See https://github.com/pallets/quart/issues/224
        # for customizing logger in the future
        app.run(
            host="0.0.0.0",
            port=args.port,
            debug=args.dev,
            use_reloader=args.dev,
        )



from __future__ import annotations

import logging
from pathlib import Path
from typing import Awaitable, Callable, Coroutine

from rich.logging import RichHandler
from hypercorn.config import Config as HyperConfig
from hypercorn.asyncio import serve
from quart import Quart as _Quart

NAME = "board"

# Use a different logger to avoid writing the logs from this plugin in
# terminal to the file (by plugin log2file)
logger = logging.getLogger(f"pipen-{NAME}")
logger.addHandler(
    RichHandler(
        show_path=False,
        omit_repeated_times=False,
        markup=True,
    )
)


class PluginNameLogFilter(logging.Filter):
    def filter(self, record):
        if isinstance(record.args, dict) and "s" in record.args:
            if record.args["s"] in (200, 304):
                record.levelno = logging.DEBUG
                record.levelname = "DEBUG"
            elif record.args["s"] == 404:
                record.levelno = logging.WARNING
                record.levelname = "WARNING"
            elif record.args["s"] >= 500:
                record.levelno = logging.ERROR
                record.levelname = "ERROR"
            if record.levelno < logger.level:
                return False

        return True


logger.addFilter(PluginNameLogFilter())
logging.getLogger("asyncio").setLevel(logging.WARNING)


# Subclass Quart to allow using logger
class Quart(_Quart):
    def run_task(
        self,
        host: str = "127.0.0.1",
        port: int = 5000,
        debug: bool | None = None,
        ca_certs: str | None = None,
        certfile: str | None = None,
        keyfile: str | None = None,
        shutdown_trigger: Callable[..., Awaitable[None]] | None = None,
    ) -> Coroutine[None, None, None]:
        config = HyperConfig()
        config.access_log_format = "%(r)s %(s)s %(b)s %(D)s"
        config.accesslog = logger
        config.bind = [f"{host}:{port}"]
        config.ca_certs = ca_certs
        config.certfile = certfile
        if debug is not None:
            self.debug = debug
        config.errorlog = config.accesslog
        config.keyfile = keyfile

        return serve(self, config, shutdown_trigger=shutdown_trigger)


# Cached/saved pipeline data
PIPEN_BOARD_DIR = Path("~/.pipen-board").expanduser()
PIPEN_BOARD_DIR.mkdir(parents=True, exist_ok=True)

SECTION_PIPELINE_OPTIONS = "PIPELINE_OPTIONS"
SECTION_ADDITIONAL_OPTIONS = "ADDITIONAL_OPTIONS"
SECTION_PROCGROUPS = "PROCGROUPS"
SECTION_PROCESSES = "PROCESSES"
SECTION_LOG = "LOG"
SECTION_DIAGRAM = "DIAGRAM"
SECTION_REPORTS = "REPORTS"

PIPELINE_OPTIONS = {
    "loglevel": {
        "type": "choice",
        "choices": ["debug", "info", "warning", "error", "critical"],
        "value": "info",
        "hidden": True,
        "desc": (
            "Logging level. " "This affects the log level of the main plugin."
        ),
    },
    "cache": {
        "type": "auto",
        "value": None,
        "placeholder": "true",
        "hidden": True,
        "desc": """# Job caching

If cache set to False (detected in the sequence of configuration files,
Pipen constructor, and process definition), the job is running anyway
regardless of previous runs.

If a previous run of a job fails, the job will be running anyway.

If a job is done successfully, a signature file will be generated for
the job. When we try to run the job again, the signature will be used
to check if we can skip running the job again but to use the results
generated by previous run.

We can also do a force-cache for a job by setting cache to "force".
This make sure of the results of previous successful run regardless of
input or script changes. This is useful for the cases that, for example,
you make some changes to input/script, but you don't want them to take
effect immediately, especially when the job takes long time to run.
""",
    },
    "submission_batch": {
        "type": "int",
        "placeholder": "8",
        "value": None,
        "hidden": True,
        "desc": "Number of jobs to submit at a time",
    },
    "scheduler": {
        "placeholder": "local",
        "value": None,
        "hidden": True,
        "desc": "The scheduler to use",
    },
    "scheduler_opts": {
        "desc": "The scheduler options",
        "value": {
            "<option_name>": {
                "type": "auto",
                "desc": "The value of the option",
                "order": 999,
            }
        },
    },
    "plugin_opts": {
        "desc": "The plugin options of your pipeline",
        "value": {
            "<plugin_name>_<plugin_opt_name>": {
                "desc": (
                    "The value of plugin option "
                    "`<plugin_name>_<plugin_opt_name>`"
                ),
                "type": "str",
                "order": 999,
            }
        },
    },
    "forks": {
        "type": "int",
        "placeholder": "1",
        "value": None,
        "desc": """Number of jobs to run in parallel for each process

The ability to run multiple jobs in parallel is provided by the the scheduler
system. For example, if you use the local scheduler, the jobs will be run in
parallel using the `multiprocessing` module. If you use the `sge` scheduler,
the jobs will be submitted to the slurm scheduler and run in parallel.
""",
    },
    "dirsig": {
        "type": "int",
        "placeholder": "1",
        "value": None,
        "hidden": True,
        "desc": "How deep we should go to check directory signature",
    },
    "error_strategy": {
        "type": "choice",
        "choices": ["ignore", "retry", "halk"],
        "choices_desc": [
            "Ignore the error and continue to run next jobs",
            "Retry the job",
            "Halt the pipeline",
        ],
        "value": "ignore",
        "desc": "What to do when a job fails",
        "hidden": True,
    },
    "num_retries": {
        "type": "int",
        "placeholder": "3",
        "value": None,
        "hidden": True,
        "desc": "Number of retries when a job fails. ",
    },
}

# See https://github.com/pwwang/xqute/blob/master/xqute/defaults.py#L34
JOB_STATUS = {
    "0": "INIT",
    "1": "RETRYING",
    "2": "QUEUED",
    "3": "SUBMITTED",
    "4": "RUNNING",
    "5": "KILLING",
    "6": "FINISHED",
    "7": "FAILED",
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pipen_board']

package_data = \
{'': ['*'], 'pipen_board': ['frontend/build/*', 'frontend/build/assets/*']}

install_requires = \
['pipen-args>=0.10.0,<0.11.0',
 'pipen-log2file>=0.3.0,<0.4.0',
 'psutil>=5.9.5,<6.0.0',
 'python-slugify>=8.0.1,<9.0.0',
 'quart>=0.18,<0.19',
 'websocket-client>=1.5,<2.0']

extras_require = \
{'diagram': ['pipen-diagram>=0.7,<0.8'],
 'report': ['pipen-report>=0.12.6,<0.13.0']}

entry_points = \
{'pipen': ['board = pipen_board:pipen_board_plugin'],
 'pipen_cli': ['cli-board = pipen_board:PipenCliBoardPlugin']}

setup_kwargs = {
    'name': 'pipen-board',
    'version': '0.9.3',
    'description': 'Visualization configuration and running of pipen pipelines on the web',
    'long_description': '# pipen-board\n\nVisualize configuration and running of [pipen][1] pipelines on the web.\n\n## Installation\n\n```bash\npip install pipen-board\n```\n\n## Usage\n\n```bash\n$ pipen board --help\nUsage: pipen board [options] <pipeline> -- [pipeline options]\n\nConfigure and run pipen pipelines from the web\n\nRequired Arguments:\n  pipeline              The pipeline and the CLI arguments to run the pipeline.\n                        For the pipeline either\n                        `/path/to/pipeline.py:<pipeline>` or\n                        `<module.submodule>:<pipeline>` `<pipeline>` must be an\n                        instance of `Pipen` and running the pipeline should be\n                        called under `__name__ == \'__main__\'.\n\nOptions:\n  -h, --help            show help message and exit\n  -p PORT, --port PORT  Port to serve the UI wizard [default: 18521]\n  -a FILE, --additional FILE\n                        Additional arguments for the pipeline, in YAML, INI,\n                        JSON or TOML format. Can have sections\n                        `ADDITIONAL_OPTIONS` and `RUNNING_OPTIONS`. It can also\n                        have other sections and items to override the\n                        configurations generated from the pipeline. If the\n                        pipeline is provided as a python script, such as\n                        `/path/to/pipeline.py:<pipeline>`, and `<pipeline>`\n                        runs under `__name__ == \'__main__\'`, the additional\n                        file can also be specified as `auto` to generate a\n                        `RUNNING OPTIONS/Local` section to run the pipeline\n                        locally.\n  --loglevel {auto,debug,info,warning,error,critical}\n                        The logging level. If `auto`, it will be set to `debug`\n                        if `--dev` is set, otherwise `info`. [default: auto]\n  --dev                 Run the pipeline in development/debug mode. This will\n                        reload the server when changes are made to this package\n                        and reload the pipeline when page reloads for new\n                        configurations. Page cache is also disabled in this\n                        mode.\n  -w WORKDIR, --workdir WORKDIR\n                        The working directory of the pipeline. [default:\n                        .pipen]\n```\n\n## Describing arguments in docstring\n\n### Docstring schema\n\n```python\nclass ProcessOrProcessGroup:\n    """Short summary\n\n    Long description\n    Long description\n\n    Args:\n        arg1 (<metadata>): description\n            - subarg1 (<metadata>): description\n            - subarg2 (<metadata>): description\n        arg2 (<metadata>): description\n\n    <Other Sections>:\n        <content>\n    """\n```\n\nThe metadata can have multiple attributes, separated by semicolon (`;`). For example:\n\n```\narg1 (action=ns;required): description\n```\n\n### Marks\n\nYou can mark a process using `pipen.utils.mark(<mark>=<value>)` as a decorator to decorate a process. For example:\n\n```python\nfrom pipen import Proc\nfrom pipen.utils import mark\n\n@mark(board_config_no_input=True)\nclass MyProc(Proc):\n    pass\n```\n\nAvailable marks:\n\n- `board_config_no_input`: Whether to show the input section for the process in configuation page. Only affects the start processes. Default to `False`.\n- `board_config_hidden`: Whether to hide the process options in the configuration page. Note that the process is still visible in the process list. Default to `False`.\n\n### Metadata for arguments\n\n\n| Name     | Description | Allowed values |\n| -------- | ----------- | -------------- |\n| `action` | Like the `action` argument in [`argx`][2]*. | `store_true`, `store_false`, `ns`, `namespace`, `append`, `extend`, `clear_append`, `clear_extend` (other values are allowed but ignore, they may be effective for CLI use) |\n| `btype`  | Board type (option type specified directly). If specified, `action` will be ignored | `ns`, `choice`, `mchoice`, `array`, `list`, `json`, `int`, `float`, `bool`, `str`, `text`, `auto`* |\n| `type` | Fallback for `action` and `btype` | Same as `btype` |\n| `flag` | Fallback for `action=store_true` | No values needed |\n| `text`/`mline`/`mlines` | Shortcut for `btype=text` | No values needed |\n| `ns`/`namespace` | Shortcut for `btype=ns` | No values needed |\n| `choices`/`choice` | Shortcut for `btype=choice` | No values needed |\n| `mchoices`/`mchoice` | Shortcut for `btype=mchoice` | No values needed |\n| `array`/`list` | Shortcut for `btype=array`/`btype=list` | No values needed |\n| `choices`/`choice` | Shortcut for `btype=choice` | No values needed |\n| `mchoices`/`mchoice` | Shortcut for `btype=mchoice` | No values needed |\n| `order` | The order of the argument in the UI. | Any integer |\n| `readonly` | Whether the argument is readonly. | No values needed (True if specified, otherwise False) |\n| `required` | Whether the argument is required. | No values needed (True if specified, otherwise False) |\n| `placeholder` | The placeholder in the UI for the argument. | Any string |\n| `bitype` | The type of the elements in an array or list. | `int`, `float`, `bool`, `str`, `json`, `auto`* |\n| `itype` | Fallback for `bitype` | Same as `bitype` |\n\n- `argx*`: An argument parser for Python, compatible with `argparse`.\n- `auto*`: Automatically infer the type from a string value.\n  - Any of `True`, `TRUE`, `true`, `False`, `FALSE`, `false` will be inferred as a `bool` value.\n  - Any of `None`, `NONE`, `none`, `null`, `NULL` will be inferred as `None`.\n  - Any integers will be inferred as `int`.\n  - Any floats will be inferred as `float`.\n  - Try to parse the value as JSON. If succeed, the value will be inferred as `json`.\n  - Otherwise, the value will be inferred as `str`.\n\n### Types of options in the UI\n\nThe type of an option in the UI is determined by the `btype`, `action` or `type` metadata. If neither is specified, a `PlainText` will be used.\n\n- `BoolOption`: Shown as a switch\n- `TextOption`: Shown as a textarea (allow multiple lines)\n- `ChoiceOption`: Shown as a dropdown list (`subarg1` and `subarg2` in the example above are used as the choices)\n- `MChoiceOption`: Shown as a multiple choice list (`subarg1` and `subarg2` in the example above are used as the choices)\n- `JsonOption`: Shown as a textarea, but the value will be validated and parsed as JSON\n- `ArrayOption`: Shown as a tag input. Items can be added or removed.\n- `AutoOption`: Shown as a 1-row textarea, and the value will be parsed automatically\n- `PlainText`: Shown as a plain text. No validation or parsing will be performed.\n- `MoreLikeOption`: Show as a box with buttons to add or remove sub-options. It\'s usally used together with `ns` type. If there is a sub-option under the option in the docstring wrapped by `<...>`, it indicates that we may have more sub-options.\n\n\n[1]: https://github.com/pwwang/pipen\n[2]: https://github.com/pwwang/argx\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from .version import __version__
from .cli import PipenCliBoardPlugin
from .plugin import PipenBoardPlugin

# Need the instance to make self work
pipen_board_plugin = PipenBoardPlugin()


def from_pipen_board():
    import sys

    return sys.argv[0] == "@pipen-board"
